import logging
from collections.abc import Iterable
from typing import Optional, Union
from types import GeneratorType

from pandas import DataFrame, RangeIndex, Series

from eda_report.exceptions import InputError


def clean_column_names(data: DataFrame) -> DataFrame:
    """Makes sure that *columns* have *meaningful* names.

    When an ``Iterable`` is used to create a ``DataFrame`` and no column names
    are provided, the columns by default are created as a
    ( :class:`~pandas.RangeIndex` ) [0, 1, 2, ...].

    This function renames such columns to ['var_1', 'var_2, 'var_3', ...],
    making references and comparisons much more intuitive.

    Parameters
    ----------
    data : DataFrame
        Data to inspect and perhaps edit.

    Returns
    -------
    DataFrame
        The data, with reader-friendly column names.
    """
    # Ensure the data has meaningful column names
    if isinstance(data.columns, RangeIndex):
        data.columns = [f"var_{i+1}" for i in data.columns]

    return data


def warn_if_target_data_has_high_cardinality(
    target_data: Series, threshold: int = 10
) -> None:
    """Check whether the ``target_data`` is suitable for color-coding or has
    too many unique values (> ``threshold``).

    Parameters
    ----------
    target_data : Series
        The data intended to color-code graphs.
    threshold : int, optional
        Maximum allowable cardinality, by default 10
    """
    if target_data.nunique() > threshold:
        logging.warning(
            f"Target variable '{target_data.name}' not used to color-code "
            "graphs since it has high cardinality "
            f"({target_data.nunique()}) which would clutter graphs."
        )


def validate_multivariate_input(data: Iterable) -> DataFrame:
    """Ensures that *multivariate input data* is of type :class:`pandas.DataFrame`.

    If it isn't, this attempts to explicitly cast it as a ``DataFrame``.

    Parameters
    ----------
    data : Iterable
        The data to analyse.

    Returns
    -------
    DataFrame
        The input data as a ``DataFrame``.

    Raises
    ------
    InputError
        If the ``data`` cannot be cast as a :class:`~pandas.DataFrame`.
    """
    if isinstance(data, DataFrame):
        return clean_column_names(data)
    else:
        try:
            data = DataFrame(data)
        except Exception:
            raise InputError(
                f"Expected a pandas.Dataframe object, but got {type(data)}."
            )
        # Attempt to infer better dtypes for object columns.
        data = data.infer_objects()
    return clean_column_names(data)


def validate_univariate_input(
    data: Iterable, *, name: Optional[str] = None
) -> Series:
    """Ensures that *univariate input data* is of type :class:`pandas.Series`.

    If it isn't, this attempts to explicitly cast it as a ``Series``.

    Parameters
    ----------
    data : Iterable
        The data to analyse.
    name : Optional[str]
        The name to assign the data, by default None.

    Returns
    -------
    Series
        The input data as a ``Series``.

    Raises
    ------
    InputError
        If the ``data`` cannot be cast as a :class:`~pandas.Series`.
    """
    if isinstance(data, GeneratorType):
        return Series(data, name=name)

    elif issubclass(type(data), Iterable) and len(list(data)) > 0:

        if isinstance(data, Series):
            name_ = name or data.name
            return data.rename(name_)

        else:
            try:
                data = Series(data, name=name)
            except Exception:
                raise InputError(
                    f"Expected a one-dimensional sequence, "
                    f"but got {type(data)}."
                )
            else:
                return data
    else:
        raise InputError("No data to process.")


def validate_target_variable(
    *, data: DataFrame, target_variable: Union[int, str]
) -> Union[Series, None]:
    """Ensures that the specified *target variable* (column label or index) is
    present in the data.

    Parameters
    ----------
    data : DataFrame
        The data being analysed.
    target_variable : Union[int, str]
        A column label or index.

    Returns
    -------
    union[Series, None]
        The target variable's data if ``target_variable`` is valid, or None.

    Raises
    ------
    InputError
        If the column label does not exist or the column index is out of
        bounds.
    """
    if target_variable is None:
        return None

    elif isinstance(target_variable, int):
        try:
            target_data = data.iloc[:, target_variable]
        except IndexError:
            raise InputError(
                f"Column index {target_variable} is not in the range"
                f" [0, {data.columns.size}]."
            )
        warn_if_target_data_has_high_cardinality(target_data)
        return target_data
    elif isinstance(target_variable, str):
        try:
            target_data = data[target_variable]
        except KeyError:
            raise InputError(
                f"{target_variable!r} is not in {data.columns.to_list()}"
            )
        warn_if_target_data_has_high_cardinality(target_data)
        return target_data
    else:
        # If target_variable is neither an index(int) or label(str)
        logging.warning(
            f"Target variable '{target_variable}' ignored."
            " Not a valid column index or label."
        )
        return None



from pathlib import Path
from typing import Union

import pandas as pd

from eda_report.exceptions import InputError


def df_from_file(filepath: Union[str, Path]) -> pd.DataFrame:
    """Reads a file, and loads its contents as a ``pandas``
    :class:`~pandas.DataFrame`.

    File formats are currently restricted to *csv* and *excel*, since these
    are the most often used to store data. This is basically a wrapper around
    ``pandas'`` input functions:

    * :func:`pandas.read_csv`
    * :func:`pandas.read_excel`

    Parameters
    ----------
    filepath : Path, str
        The path to a file.

    Returns
    -------
    :class:`~pandas.DataFrame`
        The specified file's contents.

    Raises
    ------
    InputError
        If the supplied filepath is invalid, for instance if the file is of an
        incorrect format or does not exist.
    """
    file = Path(filepath)

    if file.suffix == ".csv":
        return pd.read_csv(file)
    elif file.suffix == ".xlsx":
        return pd.read_excel(file, engine="openpyxl")
    else:
        raise InputError(f"Invalid input file: {filepath}")



import logging
from collections.abc import Iterable
from itertools import combinations
from typing import Sequence, Union

from pandas.core.frame import DataFrame

from eda_report.validate import validate_multivariate_input


class MultiVariable:
    """The defines objects that analyse data with *multiple columns*.

    Input data is held as a :class:`pandas.DataFrame` in order
    to leverage pandas_ built-in statistical methods, as well as functions
    from the `SciPy ecosystem`_.

    .. _pandas: https://pandas.pydata.org/
    .. _SciPy ecosystem: https://www.scipy.org/

    Parameters
    ----------
    data : Iterable
        The data to analyse.

    Example
    -------
    .. literalinclude:: examples.txt
       :lines: 28-58
    """

    def __init__(self, data: Iterable) -> None:
        """Initialise an instance of
        :class:`~eda_report.multivariate.MultiVariable`.
        """
        self.data = validate_multivariate_input(data)

        #: :class:`~pandas.DataFrame`: The *numeric columns* present.
        self.numeric_cols = self._select_cols("number")

        #: :class:`~pandas.DataFrame`: The *categorical columns*. Please note
        #: that **boolean** and **datetime** features are also
        #: **considered categorical** in this context.
        self.categorical_cols = self._select_cols("object", "bool")

        #: :class:`~pandas.DataFrame`: Summary statistics for numeric columns.
        self.numeric_stats = self._compute_numeric_summary_statistics()

        #: :class:`~pandas.DataFrame`: Summary statistics for categorical
        #:  columns.
        self.categorical_stats = self._compute_categorical_summary_statistics()

        #: :class:`~pandas.DataFrame`: Pearson correlation coefficients for the
        #: *numeric columns*.
        self.correlation_df = self._get_correlation()

        #: dict[tuple, str]: Brief descriptions of the nature of correlation
        #: between numeric column pairs.
        self.correlation_descriptions = {}
        self._get_bivariate_analysis()

    def __repr__(self) -> str:
        """Defines the string representation of :class:`Multivariable`.

        Returns
        -------
        str
            The string representation of the ``MultiVariable`` instance.
        """
        # Get a list of numeric features
        numeric_cols = (
            "" if self.numeric_cols is None else list(self.numeric_cols)
        )
        # Get a list of categorical features
        categoric_cols = (
            "" if self.categorical_cols is None else (self.categorical_cols)
        )
        correlation_description = (
            self._corr_description
            if hasattr(self, "_corr_description")
            else "N/A"
        )
        return "\n".join(
            [
                "\t\t\tOVERVIEW",
                "\t\t\t========",
                f"Numeric features: {', '.join(numeric_cols)}",
                f"Categorical features: {', '.join(categoric_cols)}",
                "\t\t\t  ***",
                "\t  Summary Statistics (Numeric features)",
                "\t  -------------------------------------",
                f"{self.numeric_stats if any(numeric_cols) else 'N/A'}",
                "\t\t\t  ***",
                "\t  Summary Statistics (Categorical features)",
                "\t  -----------------------------------------",
                f"{self.categorical_stats if any(categoric_cols) else 'N/A'}",
                "\t\t\t  ***",
                "\t  Bivariate Analysis (Correlation)",
                "\t  --------------------------------",
                f"{correlation_description}",
            ]
        )

    def describe(self) -> None:
        """Display summary statistics for both numerical and categorical
        columns.
        """
        if self.numeric_stats is None:
            print("\nThere are no numeric columns.\n")
        else:
            print(
                "\n\tSummary Statistics (Numeric columns):\n\n",
                self.numeric_stats,
            )

        if self.categorical_stats is None:
            print("\nThere are no categorical columns.\n")
        else:
            print(
                "\n\tSummary Statistics (Categorical columns):\n\n",
                self.categorical_stats,
            )

    def _select_cols(self, *dtypes: Sequence[str]) -> Union[DataFrame, None]:
        """Get a DataFrame including only the specified ``dtypes``.

        Returns
        -------
        Union[DataFrame, None]
            A dataframe with the desired data types.
        """
        selected_cols = self.data.select_dtypes(include=dtypes)
        return selected_cols if selected_cols.shape[1] > 0 else None

    def _compute_numeric_summary_statistics(self) -> DataFrame:
        """Get descriptive statistics for numeric columns.

        Returns
        -------
        DataFrame
            Numeric summary statistics.
        """
        if self.numeric_cols is not None:
            numeric_stats = self.numeric_cols.describe().T
            numeric_stats["skewness"] = self.numeric_cols.skew(
                numeric_only=True
            )
            numeric_stats["kurtosis"] = self.numeric_cols.kurt(
                numeric_only=True
            )

            return numeric_stats.round(4)
        else:
            return None

    def _compute_categorical_summary_statistics(self) -> DataFrame:
        """Get summary descriptive statistics for categorical columns.

        Returns
        -------
        DataFrame
            Categorical summary statistics.
        """
        if self.categorical_cols is not None:
            categorical_stats = self.categorical_cols.describe().T
            categorical_stats["relative freq"] = (
                categorical_stats["freq"] / len(self.data)
            ).apply(lambda x: f"{x :.2%}")
            return categorical_stats
        else:
            return None

    def _get_correlation(self) -> Union[DataFrame, None]:
        """Get a DataFrame of the correlation coefficients for numeric
        columns.

        Returns
        -------
        Union[DataFrame, None]
            A dataframe of Pearson correlation coefficients, or None.
        """
        return None if self.numeric_cols is None else self.numeric_cols.corr()

    def _get_variable_pairs(self) -> None:
        """Get a list of unique pairings of the numeric columns."""
        self.var_pairs = set(combinations(self.correlation_df.columns, r=2))

    def _quantify_correlation(self, var1: str, var2: str) -> None:
        """Explain the nature and magnitude of correlation between column
        pairs.

        Parameters
        ----------
        var1, var2 : str
            Numeric column labels.
        """
        correlation = self.correlation_df.loc[var1, var2]
        nature = " positive" if correlation > 0 else " negative"

        value = abs(correlation)
        if value >= 0.9:
            strength = "very strong"
        elif value >= 0.7:
            strength = "strong"
        elif value >= 0.5:
            strength = "moderate"
        elif value >= 0.3:
            strength = "weak"
        elif value >= 0.1:
            strength = "very weak"
        else:
            strength = "virtually no"
            nature = ""

        self.correlation_descriptions[
            (var1, var2)
        ] = f"{strength}{ nature} correlation ({correlation:.2f})"

    def _compare_variable_pairs(self) -> None:
        """Get a brief summary of the nature of correlation between pairs of
        numeric columns.
        """
        self._get_variable_pairs()

        for var1, var2 in self.var_pairs:
            self._quantify_correlation(var1, var2)

    def _get_bivariate_analysis(self) -> None:
        """Compare numeric column pairs."""
        if self.numeric_cols is not None and self.numeric_cols.shape[1] > 1:
            self._compare_variable_pairs()
            self._corr_description = "\n".join(
                [
                    f"{var_pair[0]} & {var_pair[1]} --> {corr_description}"
                    for var_pair, corr_description in sorted(
                        self.correlation_descriptions.items()
                    )
                ]
            )
        else:
            logging.warning(
                "Skipped Bivariate Analysis: "
                "Not enough numeric variables to compare."
            )
            return None



import logging
from typing import Iterable, Optional, Sequence, Union

from docx import Document
from docx.shared import Inches, Pt
from pandas.core.frame import DataFrame
from tqdm import tqdm

from eda_report.multivariate import MultiVariable
from eda_report.plotting import PlotMultiVariate, PlotUnivariate
from eda_report.univariate import Variable
from eda_report.validate import validate_target_variable

logging.basicConfig(
    format="[%(levelname)s %(asctime)s.%(msecs)03d] %(message)s",
    level=logging.INFO,
    datefmt="%H:%M:%S",
)


class ReportContent:
    """This defines objects that analyse data and store results as textual
    summaries and graphs.

    Parameters
    ----------
    data : Iterable
        The data to analyse.
    target_variable : Optional[Union[str, int]]
        The column
    title : str, optional
        The title to assign the report, by default "Exploratory Data Analysis
        Report"
    graph_color : str, optional
        The color to apply to the graphs, by default "cyan"
    """

    def __init__(
        self,
        data: Iterable,
        *,
        title: str = "Exploratory Data Analysis Report",
        graph_color: str = "cyan",
        target_variable: Optional[Union[str, int]] = None,
    ) -> None:
        self.variables = MultiVariable(data)
        self.GRAPH_COLOR = graph_color
        self.TARGET_VARIABLE = validate_target_variable(
            data=self.variables.data, target_variable=target_variable
        )
        self.TITLE = title
        self.intro_text = self._get_introductory_summary()
        self.multivariate_graphs = PlotMultiVariate(
            self.variables,
            graph_color=self.GRAPH_COLOR,
            hue=self.TARGET_VARIABLE,
        ).graphs
        self.variable_descriptions = self._get_variable_descriptions()
        self.bivariate_summaries = self._get_bivariate_summaries()

    def _get_introductory_summary(self) -> str:
        """Get a brief overview of the number of rows and the nature of
        columns.

        Returns
        -------
        str
            An overview of the input data.
        """
        num_rows, num_cols = self.variables.data.shape

        if num_rows == 1:
            rows = "1 row (observation)"
        else:
            rows = f"{num_rows} rows (observations)"

        if num_cols == 1:
            cols = "1 column (feature)"
        else:
            cols = f"{num_cols} columns (features)"

        # Get numeric column info
        if self.variables.numeric_cols is None:
            numeric = ""
        elif self.variables.numeric_cols.shape[1] == 1:
            numeric = ", 1 of which is numeric"
        else:
            numeric = (
                f", {self.variables.numeric_cols.shape[1]} of which are"
                " numeric"
            )

        return f"The dataset consists of {rows} and {cols}{numeric}."

    def _describe_variable(self, name: str) -> dict:

        # Perform univariate analysis using the Variable class
        var = Variable(self.variables.data[name])

        if var.num_unique == 1:
            unique_vals = "1 unique value"
        else:
            unique_vals = f"{var.num_unique} unique values"

        return {
            "description": (
                f"{name.capitalize()} is a {var.var_type} variable with"
                f" {unique_vals}. {var.missing} of its values are missing."
            ),
            "graphs": PlotUnivariate(
                variable=var,
                graph_color=self.GRAPH_COLOR,
                hue=self.TARGET_VARIABLE,
            ).graphs,
            "statistics": var.statistics,
            "most_common_items": (
                var.most_common_items
                if hasattr(var, "most_common_items")
                else None
            ),
        }

    def _get_variable_descriptions(self) -> dict:
        """Get brief descriptions of all columns present in the data.

        Returns
        -------
        dict[str, str]
            Summaries of columns present.
        """
        return {
            col: self._describe_variable(col)
            for col in tqdm(
                self.variables.data.columns,
                bar_format="{desc}: {percentage:3.0f}%|{bar:35}| "
                + "{n_fmt}/{total_fmt} features.",
                desc="Univariate analysis",
                dynamic_ncols=True,
            )
        }

    def _get_bivariate_summaries(self) -> dict:
        """Get descriptions of the nature of correlation between numeric
        column pairs.

        Returns
        -------
        dict[str, str]
            Correlation info.
        """
        if hasattr(self.variables, "var_pairs"):
            return {
                var_pair: (
                    f"{var_pair[0].title()} and {var_pair[1].title()} have"
                    f" {self.variables.correlation_descriptions[var_pair]}."
                )
                for var_pair in self.variables.var_pairs
            }
        else:
            return None


class ReportDocument(ReportContent):
    """This defines objects that produce *Word* documents with analysis
    results.

    The :class:`ReportContent` class is used to analyse the data and
    generate text and graph content. Then the :class:`~docx.document.Document`
    class from the `python-docx`_ package is used to publish the results as a
    *Word* document.

    The report is consists of 3 main sections:

    #. An *Overview* of the data and its features.
    #. *Univariate Analysis*: Summary statistics and graphs for each
        feature.
    #. *Bivariate Analysis*: Pairwise comparisons of all numerical features.

    .. _python-docx: https://python-docx.readthedocs.io/en/latest/

    Parameters
    ----------
    data : Iterable
        The data to analyse.
    target_variable : Optional[Union[str, int]]
        The column
    title : str, optional
        The title to assign the report, by default "Exploratory Data Analysis
        Report"
    graph_color : str, optional
        The color to apply to the graphs, by default "cyan"
    output_filename : str, optional
        The file name or path to save the document to, by default
        "eda-report.docx"
    table_style : str, optional
        The style to apply to the tables created, by default "Table Grid"
    """

    def __init__(
        self,
        data: Iterable,
        *,
        title: str = "Exploratory Data Analysis Report",
        graph_color: str = "cyan",
        target_variable: Optional[Union[str, int]] = None,
        output_filename: str = "eda-report.docx",
        table_style: str = "Table Grid",
    ) -> None:
        super().__init__(
            data,
            title=title,
            graph_color=graph_color,
            target_variable=target_variable,
        )
        self.OUTPUT_FILENAME = output_filename
        self.TABLE_STYLE = table_style
        self.document = Document()
        self._get_report()

    def _get_report(self):
        """Calculate summary statistics, plot graphs, and save the results as
        a .docx file.
        """
        self._create_title_page()
        self._get_variable_info()

        if hasattr(self.variables, "var_pairs"):
            self._get_bivariate_analysis()

        self._save_file()
        logging.info(f"Done. Results saved as {self.OUTPUT_FILENAME!r}")

    def _create_title_page(self) -> None:
        """Add a title and a brief summary of the data."""
        self.document.add_heading(self.TITLE, level=0)

        self.document.add_paragraph(self.intro_text)
        self.document.add_paragraph()

        self._get_numeric_overview_table()
        self._get_categorical_overview_table()

    def _get_numeric_overview_table(self) -> None:
        """Create a table with an overview of the numeric features present."""
        if self.variables.numeric_cols is not None:
            self.document.add_heading("Overview of Numeric Features", level=2)
            self.document.add_paragraph()

            # count | mean | std | min | 25% | 50% | 75% | max
            self._create_table(
                data=self.variables.numeric_stats,
                header=True,
                column_widths=(1.2,) + (0.7,) * 8,
                font_size=8.5,
                style="Normal Table",
            )

    def _get_categorical_overview_table(self) -> None:
        """Create a table with an overview of the categorical features
        present.
        """
        if self.variables.categorical_cols is not None:
            self.document.add_heading(
                "Overview of Categorical Features", level=2
            )
            self.document.add_paragraph()

            # column-name | count | unique | top | freq | relative freq
            self._create_table(
                data=self.variables.categorical_stats,
                header=True,
                column_widths=(1.2,) + (0.9,) * 5,
                font_size=8.5,
                style="Normal Table",
            )

    def _get_variable_info(self) -> None:
        """Get a brief introduction, summary statistics, and graphs for each
        individual variable.
        """
        self.document.add_heading("A. Univariate Analysis", level=1)

        for idx, var_name in enumerate(self.variable_descriptions, start=1):
            var_info = self.variable_descriptions[var_name]

            self.document.add_heading(f"{idx}. {var_name}".title(), level=2)
            self.document.add_paragraph(var_info["description"])

            self.document.add_heading("Summary Statistics", level=4)
            self._create_table(var_info["statistics"], column_widths=[2.5, 2])

            # Add most-common-items table, if present
            if var_info.get("most_common_items") is not None:
                self.document.add_heading("Most Common Values", level=4)
                self._create_table(
                    var_info["most_common_items"], column_widths=(2.5, 2)
                )

            for graph in var_info["graphs"].values():
                self.document.add_picture(graph, width=Inches(5.4))

        self.document.add_page_break()

    def _get_bivariate_analysis(self) -> None:
        """Get comparisons, scatterplots and ecdf plots for pairs of numeric
        variables.
        """
        bivariate_heading = self.document.add_heading(
            "B. Bivariate Analysis (Correlation)", level=1
        )
        bivariate_heading.paragraph_format.space_before = Pt(0)
        self.document.add_paragraph()

        self.document.add_picture(
            self.multivariate_graphs["correlation_heatmap"], width=Inches(6.7)
        )
        self.document.add_page_break()

        for idx, var_pair in enumerate(self.bivariate_summaries, start=1):
            self.document.add_heading(
                f"{idx}. {var_pair[0]} vs {var_pair[1]}".title(), level=2
            )
            self.document.add_paragraph(self.bivariate_summaries[var_pair])
            self.document.add_picture(
                self.multivariate_graphs["scatterplots"][var_pair],
                width=Inches(6),
            )
            self.document.add_paragraph()

            # Add a page break after every 2 pairs
            if idx % 2 == 0:
                self.document.add_page_break()

    def _create_table(
        self,
        data: DataFrame,
        column_widths: Sequence = (),
        font_face: str = "Courier New",
        font_size: float = 10,
        style: Optional[str] = None,
        header: bool = False,
    ) -> None:
        """Generates a table for the supplied `data` with

        Parameters
        ----------
        data : DataFrame
            The data to tabulate.
        column_widths : Sequence, optional
            The desired number and widths of columns, by default ().
        font_face : str, optional
            The font typeface for cell text, by default "Courier New".
        font_size : float, optional
            The font size for cell text, by default 10.
        style : Optional[str], optional
            A Word table style, by default None.
        header : bool, optional
            Flags whether the first row is a header, by default False.
        """
        if header:
            data.loc["", :] = data.columns
            # Sort the index in ascending order, to ensure that column names
            # end up as the first row.
            data.sort_index(inplace=True)

        table = self.document.add_table(
            rows=len(data),
            cols=len(column_widths),
        )
        table.style = style or self.document.styles[self.TABLE_STYLE]

        # Set column dimensions
        for idx, width in enumerate(column_widths):
            table.columns[idx].width = Inches(width)

        # Populate the rows
        for idx, row_data in enumerate(data.itertuples()):
            for cell, value in zip(table.rows[idx].cells, row_data):
                cell.text = f"{value}"

                # Font size and type-face have to be set at `run` level
                run = cell.paragraphs[0].runs[0]
                font = run.font
                font.size = Pt(font_size)
                font.name = font_face

        self.document.add_paragraph()

    def _save_file(self) -> None:
        """Save the document as a file."""
        self.document.save(self.OUTPUT_FILENAME)



from io import BytesIO
from typing import Optional

import matplotlib
import numpy as np
import seaborn as sns
from matplotlib.figure import Figure
from pandas import Series
from scipy.stats import probplot
from tqdm import tqdm

from eda_report.multivariate import MultiVariable
from eda_report.univariate import Variable

# Matplotlib configuration
matplotlib.rc("figure", autolayout=True, dpi=150, figsize=(6, 3.8))
matplotlib.rc("savefig")
matplotlib.rc("font", family="serif")
matplotlib.rc("axes.spines", top=False, right=False)
matplotlib.rc("axes", titlesize=12)
matplotlib.use("agg")  # use non-interactive matplotlib back-end


def savefig(figure: Figure) -> BytesIO:
    """Saves the contents of a :class:`~matplotlib.figure.Figure` in PNG format,
    as bytes in a file-like object.

    This is a utility function helpful in by-passing the *filesystem*. Graphs
    are stored in :class:`~io.BytesIO` objects, and can then be read
    directly as *attributes*, thus allowing rapid in-memory access.

    Parameters
    ----------
    figure : Figure
        A matplotlib figure with graph content.

    Returns
    -------
    BytesIO
        A graph in PNG format as bytes.
    """
    graph = BytesIO()
    figure.savefig(graph, format="png")

    return graph


class BasePlot:
    """The base class for plotting objects.

    This is where general plot settings such as the the color palette and hue
    are set.

    Parameters
    ----------
    graph_color : str, optional
        The color to apply to the generated graphs, by default "cyan".
    hue : Optional[Series]
        Data to use to group values & color-code graphs, by default None.
    """

    def __init__(
        self, *, graph_color: str = "cyan", hue: Optional[Series] = None
    ) -> None:
        self.GRAPH_COLOR = graph_color
        self.HUE = hue
        sns.set_palette(
            f"dark:{self.GRAPH_COLOR}_r",
            n_colors=2 if self.HUE is None else self.HUE.nunique(),
        )

        #: bool: A flag that determines whether or not to use the supplied
        #: ``hue``. True if ``hue`` has manageable cardinality, False
        #: otherwise.
        self.COLOR_CODE_GRAPHS = (
            True
            if self.HUE is not None and self.HUE.nunique() in range(1, 11)
            else False
        )
        self._COLOR_CODED_GRAPHS = set()


class PlotUnivariate(BasePlot):
    """This defines objects that plot instances of
    :class:`~eda_report.univariate.Variable`, which have one-dimensional data.

    Parameters
    ----------
    variable : Variable
        The data to plot.
    graph_color : str, optional
        The color to apply to the generated graphs, by default "cyan".
    hue : Optional[Series]
        Data to use to group values & color-code graphs, by default None.

    Attributes
    ----------
    graphs : Dict[str, BytesIO]
        A dictionary of graphs, with graph names as keys and file-objects
        containing plotted graphs as values.
    """

    def __init__(
        self,
        variable: Variable,
        *,
        graph_color: str = "cyan",
        hue: Optional[Series] = None,
    ) -> None:
        super().__init__(graph_color=graph_color, hue=hue)
        self.variable = variable
        self._plot_graphs()

    def _plot_graphs(self) -> None:
        """Plot graphs based on the ``Variable`` type.

        For **numeric** variables, a *box-plot*, *histogram*, *probability
        plot* and *run plot* are produced.

        For **categorical**, **boolean** or **datetime** objects, only a *bar
        plot* is produced.
        """
        if self.variable.var_type == "numeric":
            self.graphs = {
                "boxplot": self._plot_boxplot(),
                "histogram": self._plot_histogram(),
                "prob_plot": self._plot_prob_plot(),
                "run_plot": self._plot_run_plot(),
            }
        else:  # {"boolean", "categorical", "datetime"}:
            self.graphs = {"bar_plot": self._plot_bar()}

    def _plot_boxplot(self) -> BytesIO:
        """Get a boxplot for a numeric variable.

        Returns
        -------
        BytesIO
            The boxplot in PNG format as bytes in a file-object.
        """
        fig = Figure()
        ax = fig.subplots()

        if self.COLOR_CODE_GRAPHS:
            sns.boxplot(y=self.variable.data, x=self.HUE, ax=ax)
            self._COLOR_CODED_GRAPHS.add("boxplot")
        else:
            ax.boxplot(self.variable.data.dropna(), vert=False, notch=True)
            ax.set_yticklabels([""])  # Remove y-tick label
            ax.set_xlabel(f"{self.variable.name}")

        ax.set_title(f"Box-plot of {self.variable.name}")
        return savefig(fig)

    def _plot_histogram(self) -> BytesIO:
        """Get a histogram for a numeric variable.

        Returns
        -------
        BytesIO
            The histogram in PNG format as bytes in a file-object.
        """
        fig = Figure()
        ax = fig.subplots()

        if self.COLOR_CODE_GRAPHS:
            hue = self.HUE
            self._COLOR_CODED_GRAPHS.add("histogram")
        else:
            hue = None

        sns.histplot(
            x=self.variable.data,
            hue=hue,
            kde=True,
            ax=ax,
            palette=f"dark:{self.GRAPH_COLOR}_r",
        )

        ax.set_title(f"Distribution plot of {self.variable.name}")
        return savefig(fig)

    def _plot_prob_plot(self) -> BytesIO:
        """Get a probability plot for a numeric variable.

        Returns
        -------
        BytesIO
            The probability plot in PNG format as bytes in a file-object.
        """
        fig = Figure()
        ax = fig.subplots()

        theoretical_quantiles, ordered_values = probplot(
            self.variable.data,
            fit=False,  # The line of best fit will be plotted in regplot
        )
        sns.regplot(x=theoretical_quantiles, y=ordered_values, ax=ax)

        ax.set_title(f"Probability Plot of {self.variable.name}")
        ax.set_xlabel("Theoretical Quantiles (~ Standard Normal)")
        ax.set_ylabel("Ordered Values")
        return savefig(fig)

    def _plot_run_plot(self) -> BytesIO:
        """Get a run plot for a numeric variable.

        Returns
        -------
        BytesIO
            The run plot in PNG format as bytes in a file-object.
        """
        fig = Figure()
        ax = fig.subplots()

        if self.COLOR_CODE_GRAPHS:
            sns.lineplot(
                x=self.variable.data.index,
                y=self.variable.data,
                hue=self.HUE,
                ax=ax,
                palette=f"dark:{self.GRAPH_COLOR}_r",
            )
            self._COLOR_CODED_GRAPHS.add("run-plot")
        else:
            ax.plot(self.variable.data, marker=".", color=self.GRAPH_COLOR)

        ax.tick_params(axis="x", rotation=45)
        ax.set_title(f"Run Plot of {self.variable.name}")
        ax.set_ylabel("Observed Value")
        ax.set_xlabel("Index")

        # Get x-axis boundaries
        xmin, xmax = self.variable.data.index[[0, -1]]

        # Get percentiles and plot them as horizontal lines
        p5 = self.variable.data.quantile(0.05)
        p50 = self.variable.data.quantile(0.5)
        p95 = self.variable.data.quantile(0.95)

        ax.hlines(p5, xmin, xmax, "grey", "--")
        ax.text(xmax, p5, " $5^{th}$ Percentile")

        ax.hlines(p50, xmin, xmax, "grey", "--")
        ax.text(xmax, p50, " Median")

        ax.hlines(p95, xmin, xmax, "grey", "--")
        ax.text(xmax, p95, " $95^{th}$ Percentile")

        return savefig(fig)

    def _plot_bar(self) -> BytesIO:
        """Get a barplot for a categorical, boolean or datetime variable.

        Returns
        -------
        BytesIO
            The barplot in PNG format as bytes in a file-object.
        """
        fig = Figure(figsize=(6, 4), linewidth=1)
        ax = fig.subplots()

        # Include no more than 10 of the most common values
        top_10 = self.variable.data.value_counts().nlargest(10)

        if self.COLOR_CODE_GRAPHS:
            data_with_hue = (
                self.variable.data.to_frame(name="data")
                .assign(_hue_=self.HUE.values)
                .query("data in @top_10.index")
            )
            sns.countplot(data=data_with_hue, x="data", hue="_hue_", ax=ax)
            self._COLOR_CODED_GRAPHS.add("bar_plot")
        else:

            top_10 = self.variable.data.value_counts().nlargest(10)
            ax.bar(top_10.index.to_list(), top_10)

        ax.tick_params(axis="x", rotation=45)

        if self.variable.num_unique > 10:
            ax.set_title(
                f"Bar-plot of {self.variable.name} (Top 10 of "
                f"{self.variable.num_unique})"
            )
        else:
            ax.set_title(f"Bar-plot of {self.variable.name}")

        # Annotate bars
        for p in ax.patches:
            ax.annotate(
                f"{p.get_height():,}",
                ha="left",
                xy=(p.get_x(), p.get_height() * 1.02),
            )

        return savefig(fig)


class PlotMultiVariate(BasePlot):
    """This defines objects that plot instances of
    :class:`~eda_report.multivariate.MultiVariable`, which have
    two-dimensional data.

    Parameters
    ----------
    variable : MultiVariable
        The data to plot.
    graph_color : str, optional
        The color to apply to the generated graphs, by default "cyan".
    hue : Optional[Series]
        Data to use to group values & color-code graphs, by default None.

    Attributes
    ----------
    graphs : Dict[str, Union[BytesIO, dict[tuple(str, str), BytesIO]]]
        A dictionary of graphs, with graph names as keys, and file-objects
        containing the plotted graphs as values.

        Bi-variate scatterplots are further nested in a dict of
        :class:`~io.BytesIO` objects, with tuples (col_i, col_j) as keys.
    """

    def __init__(
        self,
        multivariable: MultiVariable,
        *,
        graph_color: str = "cyan",
        hue: Optional[Series] = None,
    ) -> None:
        super().__init__(graph_color=graph_color, hue=hue)
        self.multivariable = multivariable
        self._plot_graphs()

    def _plot_graphs(self) -> None:
        """Get a heatmap of the correlation in all numeric columns, and
        scatter-plots & ecdf-plots of numeric column pairs.
        """
        if hasattr(self.multivariable, "var_pairs"):
            self.bivariate_scatterplots = {
                (var_pair): self._regression_plot(*var_pair)
                for var_pair in tqdm(
                    self.multivariable.var_pairs,
                    bar_format="{desc}: {percentage:3.0f}%|{bar:35}| "
                    + "{n_fmt}/{total_fmt} numeric pairs.",
                    dynamic_ncols=True,
                    desc="Bivariate analysis",
                )
            }
            self.graphs = {
                "correlation_heatmap": self._plot_correlation_heatmap(),
                "scatterplots": self.bivariate_scatterplots,
            }
        else:
            self.graphs = None

    def _plot_correlation_heatmap(self) -> BytesIO:
        """Get a heatmap of the correlation among all numeric columns.

        Returns
        -------
        BytesIO
            The heatmap in PNG format as bytes in a file-like object.
        """
        fig = Figure(figsize=(6, 6))
        ax = fig.subplots()

        sns.heatmap(
            self.multivariable.correlation_df,
            annot=True,
            yticklabels=True,
            mask=np.triu(self.multivariable.correlation_df),
            ax=ax,
            cmap=sns.light_palette(self.GRAPH_COLOR, as_cmap=True),
        )
        ax.tick_params(rotation=45)
        fig.suptitle("Correlation in Numeric Columns", size=15)

        return savefig(fig)

    def _regression_plot(self, var1: str, var2: str) -> BytesIO:
        """Get a scatter-plot and ecdf-plot for the provided numeric columns.

        Parameters
        ----------
        var1, var2 : str
            A numeric column label.

        Returns
        -------
        BytesIO
            The scatter-plot and ecdf-plot (subplots) in PNG format as bytes
            in a file-like object.
        """
        fig = Figure(figsize=(8.2, 4))
        ax1, ax2 = fig.subplots(nrows=1, ncols=2)

        sns.regplot(x=var1, y=var2, data=self.multivariable.data, ax=ax1)

        sns.ecdfplot(
            data=self.multivariable.data.loc[:, [var1, var2]],
            ax=ax2,
            palette=f"dark:{self.GRAPH_COLOR}_r",
        )
        ax1.set_title(f"Scatter-plot - {var1} vs {var2}".title(), size=9)
        ax2.set_title("Empirical Cummulative Distribution Functions", size=9)

        return savefig(fig)



from collections.abc import Iterable
from textwrap import shorten
from typing import Optional

from pandas import DataFrame
from pandas.api.types import (is_bool_dtype, is_datetime64_any_dtype,
                              is_numeric_dtype)

from eda_report.validate import validate_univariate_input


class Variable:
    """This defines objects for analysing one-dimensional datasets — a single
    column.

    The input data is internally handled as a :class:`~pandas.Series` in order
    to leverage pandas_ built-in statistical methods, as well as functions
    from the `SciPy ecosystem`_.

    .. _pandas: https://pandas.pydata.org/
    .. _SciPy ecosystem: https://www.scipy.org/

    The result is a concise summary of the data's statistical properties.

    Parameters
    ----------
    data : Iterable
        The data to analyse.
    name : Optional[str]
        The name to assign the ``Variable``, by default None.

    Example
    --------

    .. literalinclude:: examples.txt
       :lines: 3-23
    """

    def __init__(self, data: Iterable, *, name: str = None) -> None:
        """Initialise an instance of :class:`~eda_report.univariate.Variable`.
        """
        self.data = validate_univariate_input(data, name=name)

        #: Optional[str]: The ``Variable``'s *name*. If no name is specified
        #: during instantiation, the name will be equal to the value of the
        #: ``name`` attribute of the input data (if present), or None.
        self.name = self.data.name

        #: str: The ``Variable``'s *type* — one of *"boolean"*,
        #: *"categorical"*, *"datetime"* or *"numeric"*.
        self.var_type = self._get_variable_type()

        #: :class:`~pandas.DataFrame`: The ``Variables`` *Summary statistics*.
        self.statistics = self._get_summary_statistics()

        #: int: The *number of unique values* present in the ``Variable``.
        self.num_unique = self.data.nunique()

        #: list: The *set of unique values* present in the ``Variable``.
        self.unique = sorted(self.data.dropna().unique())

        #: str: *Missing value information* in the form
        #: ``number (percentage%)``.
        self.missing = self._get_missing_values_info()

    def __repr__(self) -> str:
        """Get the string representation of the ``Variable``.

        Returns
        -------
        str
            A summary of the ``Variable``'s properties.
        """
        return "\n".join(
            [
                "\t\tOverview",
                "\t\t========",
                f"Name: {self.name}",
                f"Type: {self.var_type}",
                "Unique Values: "
                + f"{shorten(f'{self.num_unique} -> {self.unique}', 60)}",
                f"Missing Values: {self.missing}",
                "\t\t  ***",
                "\t  Summary Statistics",
                f"{self.statistics}",
            ]
        )

    def rename(self, name: Optional[str] = None) -> None:
        """Rename the ``Variable`` as specified.

        Parameters
        ----------
        name
            The name to assign to the ``Variable``, by default None.
        """
        self.name = self.data.name = name
        self.statistics.set_axis([name], axis=1, inplace=True)
        if hasattr(self, "most_common_items"):
            self.most_common_items.set_axis([name], axis=1, inplace=True)

    def _get_variable_type(self) -> str:
        """Determine the ``Variable``'s type.

        Returns
        -------
        var_type : {"boolean", "categorical", "datetime", "numeric"}
        """
        if is_numeric_dtype(self.data):
            if is_bool_dtype(self.data) or set(self.data.dropna()) == {0, 1}:
                return "boolean"
            else:
                return "numeric"
        elif is_datetime64_any_dtype(self.data):
            self.data = self.data.dt.strftime("%c")
            return "datetime"
        else:
            # Handle object, string, etc as categorical
            return "categorical"

    def _get_summary_statistics(self) -> DataFrame:
        """Get summary statistics for the column/feature."""
        if self.var_type == "numeric":
            summary = self.data.describe().set_axis(
                [
                    "Number of observations",
                    "Average",
                    "Standard Deviation",
                    "Minimum",
                    "Lower Quartile",
                    "Median",
                    "Upper Quartile",
                    "Maximum",
                ]
            )
            summary["Skewness"] = self.data.skew()
            summary["Kurtosis"] = self.data.kurt()

            return summary.round(7).to_frame()

        else:  # {"boolean", "categorical", "datetime"}
            if (self.data.shape[0] / self.data.nunique()) > 1.5:
                # If less than 2-thirds of the values are unique
                self.data = self.data.astype("category")
            else:
                self.data = self.data.astype("object")

            summary = self.data.describe()[["count", "unique", "top"]]
            summary.index = [
                "Number of observations",
                "Unique values",
                "Mode (Highest occurring value)",
            ]

            # Get most common items and their relative frequency (%)
            most_common_items = self.data.value_counts().head()
            n = len(self.data)
            self.most_common_items = most_common_items.apply(
                lambda x: f"{x} ({x / n:.2%})"
            ).to_frame()

            return summary.to_frame()

    def _get_missing_values_info(self) -> str:
        """Get the number of missing values in the ``Variable``.

        Returns
        -------
        str
            Details about the number of missing values.
        """
        missing_values = self.data.isna().sum()
        if missing_values == 0:
            return None
        else:
            return f"{missing_values} ({missing_values / len(self.data):.2%})"



import argparse

from eda_report.document import ReportDocument
from eda_report.read_file import df_from_file


def process_cli_args(*args) -> argparse.Namespace:
    """Captures and parses input from the command line interface using the
    :mod:`argparse` module from the Python standard library.

    Returns
    -------
    argparse.Namespace
        A class with parsed arguments as attributes.
    """
    parser = argparse.ArgumentParser(
        prog="eda_cli",
        description="Automatically analyse data and generate reports.",
    )

    parser.add_argument(
        "infile", type=df_from_file, help="A .csv or .xlsx file to analyse."
    )

    parser.add_argument(
        "-o",
        "--outfile",
        default="eda-report.docx",
        help="The output name for analysis results (default: %(default)s)",
    )

    parser.add_argument(
        "-t",
        "--title",
        default="Exploratory Data Analysis Report",
        help="The top level heading for the report (default: %(default)s)",
    )

    parser.add_argument(
        "-c",
        "--color",
        default="cyan",
        help="The color to apply to graphs (default: %(default)s)",
    )

    parser.add_argument(
        "-T",
        "--target",
        help=(
            "The target variable (dependent feature), used to color-code "
            "plotted values. An integer value is treated as a column index, "
            "whereas a string is treated as a column label."
            " (Default: %(default)s)"
        ),
    )

    # Parse arguments
    if args:
        return parser.parse_args(args)
    else:
        return parser.parse_args()


def run_from_cli():  # pragma: no cover
    """Creates an exploratory data analysis report in *Word* format using input
    from the command line interface.

    This is the function executed when the package is run as a script (using
    ``python -m eda_report``. It is also the entry point for the ``eda_cli``
    console script (command).

    Arguments passed from the command line are captured using the
    :func:`~eda_report.cli.process_cli_args` function, and then supplied to the
    :class:`~eda_report.document.ReportDocument` object to generate the report.
    """
    args = process_cli_args()

    ReportDocument(
        args.infile,
        title=args.title,
        graph_color=args.color,
        output_filename=args.outfile,
        target_variable=args.target,
    )



from typing import Optional, Union
from collections.abc import Iterable
from eda_report.document import ReportDocument


__version__ = "2.0.0rc0"


def get_word_report(
    data: Iterable,
    *,
    title: str = "Exploratory Data Analysis Report",
    graph_color: str = "cyan",
    target_variable: Optional[Union[str, int]] = None,
    output_filename: str = "eda-report.docx",
    table_style: str = "Table Grid",
) -> ReportDocument:
    """Analyses input data, and generates a report in *Word* (*.docx*) format.

    This is simply a wrapper around the
    :class:`~eda_report.document.ReportDocument` object, and the arguments
    supplied are passed to it.

    Parameters
    ----------
    data : Iterable
        The data to analyse.
    target_variable : Optional[Union[str, int]]
        The column
    title : str, optional
        The title to assign the report, by default "Exploratory Data Analysis
        Report"
    graph_color : str, optional
        The color to apply to the graphs, by default "cyan"
    output_filename : str, optional
        The file name or path to save the document to, by default
        "eda-report.docx"
    table_style : str, optional
        The style to apply to the tables created, by default "Table Grid"

    Returns
    -------
    ReportDocument
        The report object.
    """
    return ReportDocument(
        data,
        title=title,
        graph_color=graph_color,
        output_filename=output_filename,
        target_variable=target_variable,
        table_style=table_style
    )
